"""
Checks that files duplicated across schemes/implementations are consistent.
"""

import difflib
import os
import sys

import yaml

import helpers
import pqclean

sys.tracebacklimit = 0


def pytest_generate_tests(metafunc):
    ids = []
    argvalues = []
    if 'duplicate_consistency' not in os.environ.get('PQCLEAN_SKIP_TESTS', '').split(','):
        for scheme in pqclean.Scheme.all_schemes():
            for implementation in scheme.implementations:
                if os.path.isfile(
                        os.path.join(
                            'duplicate_consistency',
                            '{}_{}.yml'.format(scheme.name, implementation.name))):
                    metafile = os.path.join(
                        'duplicate_consistency',
                        '{}_{}.yml'.format(scheme.name, implementation.name))
                    with open(metafile, encoding='utf-8') as f:
                        metadata = yaml.safe_load(f.read())
                        for group in metadata['consistency_checks']:
                            source = pqclean.Implementation.by_name(
                                group['source']['scheme'],
                                group['source']['implementation'])
                            argvalues.append(
                                (implementation, source, group['files']))
                            ids.append(
                                "{metafile}: {scheme.name} {implementation.name}"
                                .format(scheme=scheme,
                                        implementation=implementation,
                                        metafile=metafile))
    else:
        # Placeholders so we don't crash
        ids = ['dummy']
        argvalues = [('dummy', 'dummy', 'dummy')]

    metafunc.parametrize(('implementation', 'source', 'files'),
                         argvalues,
                         ids=ids)


def file_get_contents(filename):
    with open(filename) as file:
        return file.read()


@helpers.filtered_test
def test_duplicate_consistency(implementation, source, files):
    """Test sets of files to be identical modulo namespacing"""
    messages = []
    for file in files:
        target_path = os.path.join(source.path(), file)
        this_path = os.path.join(implementation.path(), file)
        target_src = file_get_contents(target_path)\
                        .replace(source.namespace_prefix(), '')\
                        .replace(' ', '')
        this_src = file_get_contents(this_path)\
                        .replace(implementation.namespace_prefix(), '')\
                        .replace(' ', '')

        if not this_src == target_src:
            diff = difflib.unified_diff(
                this_src.splitlines(keepends=True),
                target_src.splitlines(keepends=True),
                fromfile=this_path,
                tofile=target_path)
            messages.append("{} differed:\n{}".format(file, ''.join(diff)))
    if messages:
        raise AssertionError("Files differed:\n{}".format('\n'.join(messages)))


if __name__ == '__main__':
    import pytest
    pytest.main(sys.argv)



from pathlib import Path

import helpers

@helpers.skip_windows()
def test_workflows_up_to_date():
    scriptdir = str(Path("..") / ".github" / "workflows")
    helpers.run_subprocess(["python3", "generate_workflows.py", "test"],
                           working_dir=scriptdir)

if __name__ == "__main__":
    import pytest
    import sys
    pytest.main(sys.argv)




"""
Checks that the implementation does not make use of boolean operations (==, <=, !, etc)
in assignments or function calls.
"""

import os

import pytest
import unittest

import helpers
import platform
import pqclean
import pycparser


def setup_module():
    if not(os.path.exists(os.path.join('pycparser', '.git'))):
        print("Please run `git submodule update --init`")


class ForbiddenLineVisitor(pycparser.c_ast.NodeVisitor): 
  def __init__(self):
    self.errors = []

  def visit_Assignment(self, node): 
    v = ForbiddenOpVisitor();
    v.visit(node.rvalue)
    self.errors.extend(v.errors)

  def visit_Decl(self, node):
    if node.init:
        v = ForbiddenOpVisitor();
        v.visit(node.init)
        self.errors.extend(v.errors)

  def visit_FuncCall(self, node): 
    if node.args:
        v = ForbiddenOpVisitor();
        v.visit(node.args)
        self.errors.extend(v.errors)

class ForbiddenOpVisitor(pycparser.c_ast.NodeVisitor): 
  def __init__(self):
    self.errors = []

  def visit_BinaryOp(self, node): 
    v = ForbiddenOpVisitor();
    v.visit(node.left)
    self.errors.extend(v.errors)
    if node.op in ['<', '<=', '>', '>=', '==', '!=', '&&', '||']:
      err = "\n {} at {c.file}:{c.line}:{c.column}".format(node.op, c=node.coord)
      self.errors.append(err)
    v = ForbiddenOpVisitor();
    v.visit(node.right)
    self.errors.extend(v.errors)

  def visit_UnaryOp(self, node): 
    if node.op == '!':
      err = "\n {} at {c.file}:{c.line}:{c.column}".format(node.op, c=node.coord)
      self.errors.append(err)
    v = ForbiddenOpVisitor();
    v.visit(node.expr)
    self.errors.extend(v.errors)

  def visit_TernaryOp(self, node): 
    err = "\n ternary operator at {c.file}:{c.line}:{c.column}".format(c=node.coord)
    self.errors.append(err)


@pytest.mark.parametrize(
    'implementation',
    pqclean.Scheme.all_implementations(),
    ids=str,
)
@helpers.skip_windows()
@helpers.filtered_test
def test_boolean(implementation):
    errors = []
    # pyparser's fake_libc does not support the Arm headers
    if 'supported_platforms' in implementation.metadata():
        for platform in implementation.metadata()['supported_platforms']:
            if platform['architecture'] == "arm_8":
                raise unittest.SkipTest()

    for fname in os.listdir(implementation.path()):
        if not fname.endswith(".c"):
            continue
        tdir, _ = os.path.split(os.path.realpath(__file__))
        ast = pycparser.parse_file(
            os.path.join(implementation.path(), fname),
            use_cpp=True,
            cpp_path='cc',  # not all platforms link cpp correctly; cc -E works
            cpp_args=[
                '-E',
                '-std=c99',
                '-nostdinc',  # pycparser cannot deal with e.g. __attribute__
                '-I{}'.format(os.path.join(tdir, "common")),  # include test common files to overrule compat.h
                '-I{}'.format(os.path.join(tdir, "../common")),
                # necessary to mock e.g. <stdint.h>
                '-I{}'.format(
                    os.path.join(tdir, 'pycparser/utils/fake_libc_include')),
            ]
        )
        v = ForbiddenLineVisitor()
        v.visit(ast)
        errors.extend(v.errors)
    if errors:
        raise AssertionError(
            "Prohibited use of boolean operations in assignment or function call" +
            "".join(errors)
        )


if __name__ == "__main__":
    import sys
    pytest.main(sys.argv)



"""
Checks that the functional test program (functest) can be successfully built
and executed for every scheme/implementation.
"""

import os
import platform
import unittest

import pytest

import helpers
import pqclean


@pytest.mark.parametrize(
    'implementation,test_dir,impl_path, init, destr',
    [(impl, *helpers.isolate_test_files(impl.path(), 'test_functest_'))
     for impl in pqclean.Scheme.all_supported_implementations()],
    ids=[str(impl) for impl in pqclean.Scheme.all_supported_implementations()],
)
@helpers.filtered_test
def test_functest(implementation, impl_path, test_dir,
                  init, destr):
    init()
    dest_dir = os.path.join(test_dir, 'bin')
    helpers.make('functest',
                 TYPE=implementation.scheme.type,
                 SCHEME=implementation.scheme.name,
                 IMPLEMENTATION=implementation.name,
                 SCHEME_DIR=impl_path,
                 DEST_DIR=dest_dir,
                 working_dir=os.path.join(test_dir, 'test'))
    helpers.run_subprocess(
        [os.path.join(dest_dir, 'functest_{}_{}{}'.format(
            implementation.scheme.name,
            implementation.name,
            '.exe' if os.name == 'nt' else ''
        ))],
    )
    destr()


@pytest.mark.parametrize(
    'implementation,test_dir,impl_path, init, destr',
    [(impl,
      *helpers.isolate_test_files(impl.path(), 'test_functest_sanitizers_'))
     for impl in pqclean.Scheme.all_supported_implementations()],
    ids=[str(impl) for impl in pqclean.Scheme.all_supported_implementations()],
)
@helpers.skip_windows()
@helpers.filtered_test
def test_functest_sanitizers(implementation, impl_path, test_dir,
                             init, destr):
    dest_dir = os.path.join(test_dir, 'bin')
    env = None
    if (implementation.scheme.name == "sphincs-sha256-192s-robust"
            and 'CI' in os.environ
            and implementation.name == "clean"
            and 'clang' in os.environ.get('CC', '')):
        raise unittest.SkipTest("Clang makes this test use too much RAM")
    if platform.machine() == 'ppc' and 'clang' in os.environ.get('CC', 'gcc'):
        raise unittest.SkipTest("Clang does not support ASAN on ppc")
    elif platform.machine() in ['armv7l', 'aarch64']:
        env = {'ASAN_OPTIONS': 'detect_leaks=0'}
    elif platform.system() == 'Darwin':
        raise unittest.SkipTest('ASAN is not reliable on OSX')
    else:
        print("Supported platform: {}".format(platform.machine()))

    init()

    helpers.make('clean-scheme', 'functest',
                 TYPE=implementation.scheme.type,
                 SCHEME=implementation.scheme.name,
                 IMPLEMENTATION=implementation.name,
                 EXTRAFLAGS=(
                     '-g -fsanitize=address,undefined '
                     '-fno-sanitize-recover=undefined '
                     # TODO(JMS) Remove explicit -latomic if/when gcc fixes:
                     # https://gcc.gnu.org/bugzilla/show_bug.cgi?id=81358
                     '-Wno-unused-command-line-argument -latomic'),
                 SCHEME_DIR=impl_path,
                 DEST_DIR=dest_dir,
                 working_dir=os.path.join(test_dir, 'test'),
                 env=env)
    helpers.run_subprocess(
        [os.path.join(dest_dir, 'functest_{}_{}{}'.format(
            implementation.scheme.name,
            implementation.name,
            '.exe' if os.name == 'nt' else ''
        ))],
        env=env,
    )
    destr()


if __name__ == '__main__':
    import sys
    pytest.main(sys.argv)



import os
import platform
import unittest
from glob import glob

import pytest

import helpers
import pqclean

additional_flags = [] #['-fix-errors']


@pytest.mark.parametrize(
    'implementation',
    pqclean.Scheme.all_supported_implementations(),
    ids=str,
)
@helpers.skip_windows()
@helpers.filtered_test
def test_clang_tidy(implementation: pqclean.Implementation):
    helpers.ensure_available('clang-tidy')
    cfiles = implementation.cfiles()
    common_files = glob(os.path.join('..', 'common', '*.c'))
    (returncode, _) = helpers.run_subprocess(
        ['clang-tidy',
         '-quiet',
         '-header-filter=.*',
         *additional_flags,
         *cfiles,
         *common_files,
         '--',
         '-iquote', os.path.join('..', 'common'),
         '-iquote', implementation.path()],
        expected_returncode=None,
    )

    # Detect and gracefully avoid segfaults
    if returncode == -11:
        raise unittest.SkipTest("clang-tidy segfaulted")

    assert returncode == 0, "Clang-tidy returned %d" % returncode


if __name__ == "__main__":
    import sys
    pytest.main(sys.argv)



"""
Checks that (hash of the) test vectors produced on this platform matches
the one provided in the META file for every scheme/implementation.
"""

import hashlib
import os
import unittest

import pytest

import helpers
import pqclean


@pytest.mark.parametrize(
    'implementation,test_dir,impl_path,init,destr',
    [(impl, *helpers.isolate_test_files(impl.path(), 'test_testvectors_'))
     for sig in pqclean.Signature.all_sigs()
     for impl in sig.implementations],
    ids=[str(impl) for sig in pqclean.Signature.all_sigs()
         for impl in sig.implementations],
)
@helpers.filtered_test
def test_testvectors(implementation, impl_path, test_dir, init, destr):
    if not implementation.supported_on_current_platform():
        raise unittest.SkipTest("Not supported on current platform")
    init()
    dest_dir = os.path.join(test_dir, 'bin')
    helpers.make('testvectors',
                 TYPE=implementation.scheme.type,
                 SCHEME=implementation.scheme.name,
                 SCHEME_DIR=impl_path,
                 IMPLEMENTATION=implementation.name,
                 DEST_DIR=dest_dir,
                 working_dir=os.path.join(test_dir, 'test'))
    out = helpers.run_subprocess(
        [os.path.join(dest_dir, 'testvectors_{}_{}{}'.format(
            implementation.scheme.name,
            implementation.name,
            '.exe' if os.name == 'nt' else ''
        ))],
        print_output=False,
    ).replace('\r', '')
    assert(implementation.scheme.metadata()['testvectors-sha256'].lower()
           == hashlib.sha256(out.encode('utf-8')).hexdigest().lower())
    destr()


if __name__ == '__main__':
    import sys
    pytest.main(sys.argv)



"""
Runs the test files under valgrind to detect memory problems
"""

import functools
import os
import platform
import re
import unittest

import pytest

import helpers
import pqclean


@functools.lru_cache()
def valgrind_supports_exit_early():
    """Checks if we support early exit from valgrind"""
    version = helpers.run_subprocess(['valgrind', '--version'])
    match = re.match(r'valgrind-(\d)\.(\d+).*', version)
    if match:
        return int(match.group(2)) >= 14
    return False


@pytest.mark.parametrize(
    'implementation,test_dir,impl_path, init, destr',
    [(impl, *helpers.isolate_test_files(impl.path(), 'test_valgrind_'))
     for impl in pqclean.Scheme.all_supported_implementations()],
    ids=[str(impl) for impl in pqclean.Scheme.all_supported_implementations()],
)
@helpers.filtered_test
def test_valgrind(implementation: pqclean.Implementation, impl_path, test_dir,
                  init, destr):
    if (platform.machine() not in ('i386', 'x86_64', 'aarch64') or
            platform.system() != 'Linux'):
        raise unittest.SkipTest()
    init()

    dest_dir = os.path.join(test_dir, 'bin')

    helpers.make(TYPE=implementation.scheme.type,
                 SCHEME=implementation.scheme.name,
                 SCHEME_DIR=os.path.abspath(impl_path),
                 IMPLEMENTATION=implementation.name,
                 DEST_DIR=dest_dir,
                 EXTRAFLAGS="-g3",
                 NTESTS=1,
                 working_dir=os.path.join(test_dir, 'test'))
    functest_name = './functest_{}_{}'.format(implementation.scheme.name,
                                              implementation.name)
    helpers.run_subprocess(
        ['valgrind',
         '--error-exitcode=1',
         '--leak-check=yes',
         *(['--exit-on-first-error=yes']
           if valgrind_supports_exit_early()
           else []),
         '--max-stackframe=20933064',
         '--vex-guest-max-insns=25',
         functest_name],
        dest_dir)
    destr()


if __name__ == '__main__':
    import sys
    pytest.main(sys.argv)



import pytest

import helpers
import pqclean


@pytest.mark.parametrize(
    'implementation',
    pqclean.Scheme.all_implementations(),
    ids=str,
)
@helpers.filtered_test
def test_format(implementation: pqclean.Implementation):
    helpers.ensure_available('astyle')
    cfiles = implementation.cfiles()
    hfiles = implementation.hfiles()
    result = helpers.run_subprocess(
        ['astyle',
         '--dry-run',
         '--options=../.astylerc',
         *cfiles,
         *hfiles])
    assert 'Formatted' not in result


if __name__ == '__main__':
    import sys
    pytest.main(sys.argv)



"""
Verify the metadata specified in the META.yml files.
"""

import copy
import itertools

import pytest

import helpers
import pqclean


@pytest.mark.parametrize(
    'scheme',
    pqclean.Scheme.all_schemes(),
    ids=str,
)
@helpers.filtered_test
def test_metadata(scheme):
    metadata = scheme.metadata()

    specification = EXPECTED_FIELDS.items()

    if scheme.type == 'kem':
        specification = itertools.chain(specification, KEM_FIELDS.items())
    elif scheme.type == 'sign':
        specification = itertools.chain(specification,
                                        SIGNATURE_FIELDS.items())
    else:
        assert False, "Wrong type of metadata"

    check_spec(copy.deepcopy(metadata), specification)

    implementation_names_in_yaml = set(
        i['name'] for i in metadata['implementations'])
    if len(implementation_names_in_yaml) != len(metadata['implementations']):
        raise AssertionError("Implementations in YAML file are not distinct")
    implementations_on_disk = set(i.name for i in scheme.implementations)
    if implementation_names_in_yaml != implementations_on_disk:
        raise AssertionError("Implementations in YAML file {} and "
                             "implementations on disk {} do not match"
                             .format(implementation_names_in_yaml,
                                     implementations_on_disk))


EXPECTED_FIELDS = {
    'name': {'type': str},
    'type': {'type': str},
    'claimed-nist-level': {'type': int, 'min': 1, 'max': 5},
    'length-public-key': {'type': int, 'min': 1},
    'length-secret-key': {'type': int, 'min': 1},
    'nistkat-sha256': {'type': str, 'length': 64},
    'principal-submitters': {'type': list, 'elements': {'type': str}},
    'auxiliary-submitters': {
        'type': list, 'elements': {'type': str}, 'optional': True},
    'implementations': {
        'type': list,
        'elements': {
            'type': dict,
            'spec': {
                'name': {'type': str},
                'version': {'type': str},
                'supported_platforms': {
                    'type': list,
                    'optional': True,
                    'elements': {
                        'type': dict,
                        'spec': {
                            'architecture': {
                                'type': str,
                                'values': ['x86', 'x86_64', 'arm_8']},
                            'required_flags': {
                                'type': list,
                                'optional': True,
                                'elements': {'type': str},
                            },
                            'operating_systems': {
                                'type': list,
                                'optional': True,
                                'elements': {
                                    'type': str,
                                    'values': ['Linux', 'Windows', 'Darwin'],
                                },
                            },
                        },
                    },
                },
            },
        },
    },
}

KEM_FIELDS = {
    'claimed-security': {'type': str, 'values': ['IND-CPA', 'IND-CCA2']},
    'length-ciphertext': {'type': int, 'min': 1},
    'length-shared-secret': {'type': int, 'min': 1},
}

SIGNATURE_FIELDS = {
    'length-signature': {'type': int, 'min': 1},
    'testvectors-sha256': {'type': str, 'length': 64},
}


def check_spec(metadata, spec):
    for field, props in spec:
        if field not in metadata and 'optional' not in props:
            raise AssertionError("Field '{}' not present.".format(field))

        # validate element
        if field in metadata:
            check_element(field, metadata[field], props)

            # delete it to detect extras
            del metadata[field]

    # Done checking all specified fields, check if we have extras
    for field, value in metadata.items():
        raise AssertionError(
            "Unexpected item '{}' with value '{}'".format(field, value))


def check_element(field, element, props):
    type_ = props['type']
    # Validate type of element
    type_(element)

    # Strs are valid lists otherwise
    if type_ == list and type(element) != list:
        raise ValueError("Field {} not a list".format(field))

    # lists are valid dicts otherwise
    if type_ == dict and type(element) != dict:
        raise ValueError("Field {} not a dict".format(field))

    if type_ == int:
        element = int(element)
        if 'min' in props:
            if element < props['min']:
                raise ValueError("Value of field '{}' is lower than minimum "
                                 "value {}".format(field, props['min']))
        if 'max' in props:
            if element > props['max']:
                raise ValueError("Value of field '{}' is larger than maximum"
                                 " value {}"
                                 .format(field, props['max']))

    if type_ == str:
        if 'length' in props:
            actual_len = len(element)
            if actual_len != props['length']:
                raise ValueError("Value of field '{}' should be length {}"
                                 " but was length {}"
                                 .format(field, props['length'], actual_len))
        if 'values' in props and element not in props['values']:
            raise ValueError("'{}' should be in {}"
                             .format(element, props['values']))

    if type_ == list:  # recursively check the elements
        for el in element:
            check_element('element of {}'.format(field), el, props['elements'])

    if type_ == dict:
        check_spec(element, props['spec'].items())


if __name__ == '__main__':
    import sys
    pytest.main(sys.argv)



#!/usr/bin/env python3
import sys
import difflib

from pathlib import Path

import jinja2


def render_workflow(scheme_path):
    result = 0
    scheme_name = scheme_path.name
    scheme_type = str(scheme_path.parent.name).split("_")[1]
    target_name = f"{scheme_type}_{scheme_name}.yml"

    tpl = TPL_ENV.get_template("template.yml.j2")
    new_contents = tpl.render(scheme_type=scheme_type, scheme_name=scheme_name)
    if not TEST_MODE:
        with open(target_name, "w") as f:
            f.write(new_contents)
        with open("BADGES.md", "a") as f:
            f.write(
                f"![Test {scheme_name}]"
                f"(https://github.com/PQClean/PQClean/workflows/Test%20{scheme_name}/badge.svg?branch=master)\n")
    else:
        if (Path(".") / target_name).exists():
            with open(target_name, "r") as f:
                file_contents = f.read()
        else:
            file_contents = ""

        if file_contents != new_contents:
            sys.stdout.writelines(difflib.unified_diff(
                file_contents.splitlines(keepends=True), new_contents.splitlines(keepends=True),
                fromfile=f"old/{target_name}", tofile=f"new/{target_name}")
            )
            result = 1

    return result


if __name__ == "__main__":
    TEST_MODE = len(sys.argv) > 1 and sys.argv[1] == "test"

    loader = jinja2.FileSystemLoader(".")
    TPL_ENV = jinja2.Environment(
        loader=loader, variable_start_string="{-", variable_end_string="-}"
    )

    if not TEST_MODE:
        with open("BADGES.md", "w") as f:
            f.truncate()

    with open("template.yml.j2") as f:
        TEMPLATE = f.read()
    root = Path("../..")

    paths = root.glob("crypto_*/*")
    for path in paths:
        result = render_workflow(path)

    sys.exit(result)



"""
Checks that a Makefile and a Makefile.Microsoft_nmake file is present for every
implementation of the specified scheme.
"""

import os

import pytest

import helpers
import pqclean


@pytest.mark.parametrize(
    'implementation',
    pqclean.Scheme.all_implementations(),
    ids=str,
)
@helpers.filtered_test
def test_makefile_present(implementation):
    p1 = os.path.join(implementation.path(), 'Makefile')
    assert os.path.isfile(p1)


@pytest.mark.parametrize(
    'implementation',
    pqclean.Scheme.all_implementations(),
    ids=str,
)
@helpers.filtered_test
def test_microsoft_nmakefile_present(implementation):
    p2 = os.path.join(implementation.path(), 'Makefile.Microsoft_nmake')
    if implementation.supported_on_os(os='Windows'):
        assert os.path.isfile(p2)
    else:
        assert not os.path.isfile(p2), "Should not have an NMake file"


if __name__ == '__main__':
    import sys
    pytest.main(sys.argv)



"""
Checks that the all exported symbols are properly namespaced, i.e., all
start with "PQCLEAN_SCHEMENAME_".
"""

import sys
import unittest

import pytest

import helpers
import pqclean


@pytest.mark.parametrize(
    'implementation,test_dir,impl_path,init,destr',
    [(impl,
      *helpers.isolate_test_files(impl.path(), 'test_symbol_ns_'))
     for impl in pqclean.Scheme.all_supported_implementations()],
    ids=[str(impl) for impl in pqclean.Scheme.all_supported_implementations()],
)
@helpers.filtered_test
def test_symbol_namespaces(implementation, impl_path, test_dir, init, destr):
    if sys.platform not in ['linux', 'darwin']:
        raise unittest.SkipTest("Unsupported platform")
    init()
    helpers.make(working_dir=impl_path)
    out = helpers.run_subprocess(
        ['nm', '-g', implementation.libname()],
        impl_path,
    )

    lines = out.strip().split("\n")
    symbols = []
    for line in lines:
        if line.startswith("00"):
            symbols.append(line)

    namespace = implementation.namespace_prefix()
    non_namespaced = []
    for symbolstr in symbols:
        *_, symtype, symbol = symbolstr.split()
        if symtype in 'TDRS':
            if not (symbol.startswith(namespace) or
                    symbol.startswith('_' + namespace) or
                    # KeccakP-1600 for AVX2
                    symbol.startswith('KeccakF1600times4') or
                    symbol.startswith('_KeccakF1600times4') or  # MacOS
                    symbol.startswith('KeccakP1600times4') or
                    symbol.startswith('_KeccakP1600times4') or  # MacOS
                    # weird things on i386
                    symbol.startswith('__x86.get_pc_thunk.')):
                non_namespaced.append(symbol)

    if non_namespaced:
        print("Missing namespace literal {}".format(namespace))
        for symbol in non_namespaced:
            print("\ttype: {}, symbol: {}".format(symtype, symbol))

    assert not non_namespaced, "Literals with missing namespaces"
    destr()


if __name__ == '__main__':
    pytest.main(sys.argv)



import os
import re

import pytest

import helpers
import pqclean

pattern = re.compile(r'^\s*#include\s*"')


@pytest.mark.parametrize(
    'implementation',
    pqclean.Scheme.all_implementations(),
    ids=str,
)
@helpers.filtered_test
def test_api_h(implementation: pqclean.Implementation):
    apipath = os.path.join(implementation.path(), 'api.h')
    errors = []
    with open(apipath) as f:
        for i, line in enumerate(f):
            if pattern.match(line):
                errors.append("\n at {}:{}".format(apipath, i+1))
    if errors:
        raise AssertionError(
            "Prohibited external include in api.h" + "".join(errors)
        )


if __name__ == '__main__':
    import sys
    pytest.main(sys.argv)




"""
Checks that the implementation does not make use of the `char` type.
This is ambiguous; compilers can freely choose `signed` or `unsigned` char.
"""

import os

import pytest
import unittest

import helpers
import platform
import pqclean
import pycparser


def setup_module():
    if not(os.path.exists(os.path.join('pycparser', '.git'))):
        print("Please run `git submodule update --init`")


def walk_tree(ast, parent=[]):
    if type(ast) is pycparser.c_ast.IdentifierType:
        if ast.names == ['char']:
            # allow casts in certain function calls
            try:
                if parent[-5].name.name in ['_mm256_set1_epi8']:
                    return
            except:
                pass
            yield ast

    for (_, child) in ast.children():
        # recursively yield prohibited nodes
        yield from walk_tree(child, parent=parent + [ast])


@pytest.mark.parametrize(
    'implementation',
    pqclean.Scheme.all_implementations(),
    ids=str,
)
@helpers.skip_windows()
@helpers.filtered_test
def test_char(implementation):
    errors = []
    # pyparser's fake_libc does not support the Arm headers
    if 'supported_platforms' in implementation.metadata():
        for platform in implementation.metadata()['supported_platforms']:
            if platform['architecture'] == "arm_8":
                raise unittest.SkipTest()

    for fname in os.listdir(implementation.path()):
        if not fname.endswith(".c"):
            continue
        tdir, _ = os.path.split(os.path.realpath(__file__))
        ast = pycparser.parse_file(
            os.path.join(implementation.path(), fname),
            use_cpp=True,
            cpp_path='cc',  # not all platforms link cpp correctly; cc -E works
            cpp_args=[
                '-E',
                '-std=c99',
                '-nostdinc',  # pycparser cannot deal with e.g. __attribute__
                '-I{}'.format(os.path.join(tdir, "common")),  # include test common files to overrule compat.h
                '-I{}'.format(os.path.join(tdir, "../common")),
                # necessary to mock e.g. <stdint.h>
                '-I{}'.format(
                    os.path.join(tdir, 'pycparser/utils/fake_libc_include')),
            ]
        )
        for node in walk_tree(ast):
            # flatten nodes to a string to easily enforce uniqueness
            err = "\n at {c.file}:{c.line}:{c.column}".format(c=node.coord)
            if err not in errors:
                errors.append(err)
    if errors:
        raise AssertionError(
            "Prohibited use of char without explicit signed/unsigned" +
            "".join(errors)
        )


if __name__ == "__main__":
    import sys
    pytest.main(sys.argv)



import pytest

import helpers
import pqclean


@pytest.mark.parametrize(
    'implementation',
    pqclean.Scheme.all_implementations(),
    ids=str,
)
@helpers.filtered_test
def test_preprocessor(implementation: pqclean.Implementation):
    cfiles = implementation.cfiles()
    hfiles = implementation.hfiles()
    errors = []
    for file in hfiles + cfiles:
        with open(file) as f:
            for i, line in enumerate(f):
                line = line.strip()
                if file in hfiles and i == 0 and line.startswith('#ifndef'):
                    continue
                if line.startswith('#if'):
                    errors.append("\n at {}:{}".format(file, i+1))
    if errors:
        raise AssertionError(
            "Prohibited use of preprocessor conditional" + "".join(errors)
        )


if __name__ == '__main__':
    import sys
    pytest.main(sys.argv)



"""
Checks that no implementation makes use of symbolic links.
"""

import os

import pytest

import helpers
import pqclean


@pytest.mark.parametrize(
    'implementation',
    pqclean.Scheme.all_implementations(),
    ids=str,
)
@helpers.filtered_test
def test_no_symlinks(implementation):
    for file in os.listdir(implementation.path()):
        fpath = os.path.join(implementation.path(), file)
        if os.path.islink(fpath):
            raise AssertionError("{} is a symbolic link!".format(fpath))


if __name__ == '__main__':
    import sys
    pytest.main(sys.argv)



import glob
import os
from typing import Optional
from functools import lru_cache

import yaml
import platform
import helpers


class Scheme:
    def __init__(self):
        self.type = None
        self.name = None
        self.implementations = []

    def path(self, base='..'):
        return os.path.join(base, 'crypto_' + self.type, self.name)

    def namespace_prefix(self):
        return 'PQCLEAN_{}_'.format(self.name.upper()).replace('-', '')

    @staticmethod
    @lru_cache(maxsize=None)
    def by_name(scheme_name):
        for scheme in Scheme.all_schemes():
            if scheme.name == scheme_name:
                return scheme
        raise KeyError()

    @staticmethod
    @lru_cache(maxsize=1)
    def all_schemes():
        schemes = []
        schemes.extend(Scheme.all_schemes_of_type('kem'))
        schemes.extend(Scheme.all_schemes_of_type('sign'))
        return schemes

    @staticmethod
    @lru_cache(maxsize=1)
    def all_implementations():
        implementations = []
        for scheme in Scheme.all_schemes():
            implementations.extend(scheme.implementations)
        return implementations

    @staticmethod
    @lru_cache(maxsize=1)
    def all_supported_implementations():
        return [impl for impl in Scheme.all_implementations()
                if impl.supported_on_current_platform()]

    @staticmethod
    @lru_cache(maxsize=32)
    def all_schemes_of_type(type: str) -> list:
        schemes = []
        p = os.path.join('..', 'crypto_' + type)
        if os.path.isdir(p):
            for d in os.listdir(p):
                if os.path.isdir(os.path.join(p, d)):
                    if type == 'kem':
                        schemes.append(KEM(d))
                    elif type == 'sign':
                        schemes.append(Signature(d))
                    else:
                        assert('Unknown type')
        return schemes

    @lru_cache(maxsize=None)
    def metadata(self):
        metafile = os.path.join(self.path(), 'META.yml')
        try:
            with open(metafile, encoding='utf-8') as f:
                metadata = yaml.safe_load(f)
            return metadata
        except Exception as e:
            print("Can't open {}: {}".format(metafile, e))
            return None

    def __repr__(self):
        return "<{}({})>".format(self.type.title(), self.name)


class Implementation:

    def __init__(self, scheme, name):
        self.scheme = scheme
        self.name = name

    @lru_cache(maxsize=None)
    def metadata(self):
        for i in self.scheme.metadata()['implementations']:
            if i['name'] == self.name:
                return i

    def path(self, base='..') -> str:
        return os.path.join(self.scheme.path(base=base), self.name)

    def libname(self) -> str:
        if os.name == 'nt':
            return "lib{}_{}.lib".format(self.scheme.name, self.name)
        return "lib{}_{}.a".format(self.scheme.name, self.name)

    def cfiles(self) -> [str]:
        return glob.glob(os.path.join(self.path(), '*.c'))

    def hfiles(self) -> [str]:
        return glob.glob(os.path.join(self.path(), '*.h'))

    def ofiles(self) -> [str]:
        return glob.glob(os.path.join(self.path(),
                         '*.o' if os.name != 'nt' else '*.obj'))

    @staticmethod
    @lru_cache(maxsize=None)
    def by_name(scheme_name, implementation_name):
        scheme = Scheme.by_name(scheme_name)
        for implementation in scheme.implementations:
            if implementation.name == implementation_name:
                return implementation
        raise KeyError()

    @staticmethod
    @lru_cache(maxsize=None)
    def all_implementations(scheme: Scheme) -> list:
        implementations = []
        for d in os.listdir(scheme.path()):
            if os.path.isdir(os.path.join(scheme.path(), d)):
                implementations.append(Implementation(scheme, d))
        return implementations

    @staticmethod
    def all_supported_implementations(scheme: Scheme) -> list:
        return [impl for impl in Implementation.all_implementations(scheme)
                if impl.supported_on_current_platform()]

    def namespace_prefix(self):
        return '{}{}_'.format(self.scheme.namespace_prefix(),
                              self.name.upper()).replace('-', '')

    def supported_on_os(self, os: Optional[str] = None) -> bool:
        """Check if we support the OS

        If no OS is specified, then we run on the current OS
        """
        if os is None:
            os = platform.system()

        for platform_ in self.metadata().get('supported_platforms', []):
            if 'operating_systems' in platform_:
                if os not in platform_['operating_systems']:
                    return False

        return True

    @lru_cache(maxsize=10000)
    def supported_on_current_platform(self) -> bool:
        if 'supported_platforms' not in self.metadata():
            return True

        if platform.machine() == 'ppc':
            return False

        if not self.supported_on_os():
            return False

        cpuinfo = helpers.get_cpu_info()

        for platform_ in self.metadata()['supported_platforms']:
            if platform_['architecture'] == cpuinfo['arch'].lower():
                # Detect actually running on emulated i386
                if (platform_['architecture'] == 'x86_64' and
                        platform.architecture()[0] == '32bit'):
                    continue
                if all([flag in cpuinfo['flags']
                        for flag in platform_['required_flags']]):
                    return True
        return False

    def __str__(self):
        return "{} implementation of {}".format(self.name, self.scheme.name)

    def __repr__(self):
        return "<Implementation({}, {})>".format(self.scheme.name, self.name)


class KEM(Scheme):

    def __init__(self, name: str):
        self.type = 'kem'
        self.name = name
        self.implementations = Implementation.all_implementations(self)

    @staticmethod
    def all_kems() -> list:
        return Scheme.all_schemes_of_type('kem')


class Signature(Scheme):

    def __init__(self, name: str):
        self.type = 'sign'
        self.name = name
        self.implementations = Implementation.all_implementations(self)

    @staticmethod
    def all_sigs():
        return Scheme.all_schemes_of_type('sign')



"""
Checks that (hash of the) KATs (in NIST format) produced on this platform match
the one provided in the META file for every scheme/implementation.

Note that this only uses the first test case from the NIST-format KAT files.
The appropriate hash can be generated from the original submission's KAT file
using the command:
    cat PQCkemKAT_whatever.rsp | head -n 8 | tail -n 6 | sha256sum
"""

import hashlib
import os

import pytest

import helpers
import pqclean


@pytest.mark.parametrize(
    'implementation,test_dir,impl_path, init, destr',
    [(impl, *helpers.isolate_test_files(impl.path(), 'test_functest_'))
     for impl in pqclean.Scheme.all_supported_implementations()],
    ids=[str(impl) for impl in pqclean.Scheme.all_supported_implementations()],
)
@helpers.filtered_test
def test_nistkat(implementation, impl_path, test_dir, init, destr):
    init()
    dest_path = os.path.join(test_dir, 'bin')
    helpers.make('nistkat',
                 TYPE=implementation.scheme.type,
                 SCHEME=implementation.scheme.name,
                 IMPLEMENTATION=implementation.name,
                 SCHEME_DIR=impl_path,
                 DEST_DIR=dest_path,
                 working_dir=os.path.join(test_dir, 'test'))
    out = helpers.run_subprocess(
        [os.path.join(dest_path, 'nistkat_{}_{}{}'.format(
            implementation.scheme.name,
            implementation.name,
            '.exe' if os.name == 'nt' else ''
        ))],
        print_output=False,
    ).replace('\r', '')
    assert(implementation.scheme.metadata()['nistkat-sha256'].lower()
           == hashlib.sha256(out.encode('utf-8')).hexdigest().lower())
    destr()


if __name__ == '__main__':
    import sys
    pytest.main(sys.argv)



import atexit
import functools
import logging
import os
import secrets
import shutil
import string
import subprocess
import sys
import unittest
from functools import lru_cache

import pqclean


@atexit.register
def cleanup_testcases():
    """Clean up any remaining isolated test dirs"""
    print("Cleaning up testcases directory",
          file=sys.stderr)
    for dir_ in TEST_TEMPDIRS:
        shutil.rmtree(dir_, ignore_errors=True)


TEST_TEMPDIRS = []

ALPHABET = string.ascii_letters + string.digits + '_'
def mktmpdir(parent, prefix):
    """Returns a unique directory name"""
    uniq = ''.join(secrets.choice(ALPHABET) for i in range(8))
    return os.path.join(parent, "{}_{}".format(prefix, uniq))


def isolate_test_files(impl_path, test_prefix,
                       dir=os.path.join('..', 'testcases')):
    """Isolates the test files in a separate directory, to help parallelise.

    Especially Windows is problematic and needs isolation of all test files:
    its build process will create .obj files EVERYWHERE.
    """
    try:
        os.mkdir(dir)
    except FileExistsError:
        pass
    test_dir = mktmpdir(dir, test_prefix)
    test_dir = os.path.abspath(test_dir)
    TEST_TEMPDIRS.append(test_dir)

    # the implementation will go here.
    scheme_dir = os.path.join(test_dir, 'crypto_bla', 'scheme')
    new_impl_dir = os.path.abspath(os.path.join(scheme_dir, 'impl'))

    def initializer():
        """Isolate the files to be tested"""
        # Create layers in folder structure
        os.makedirs(scheme_dir)

        # Create test dependencies structure
        os.mkdir(os.path.join(test_dir, 'test'))

        # Copy common files (randombytes.c, aes.c, ...)
        shutil.copytree(
            os.path.join('..', 'common'), os.path.join(test_dir, 'common'))
        # Copy makefiles
        shutil.copy(os.path.join('..', 'test', 'Makefile'),
                    os.path.join(test_dir, 'test', 'Makefile'))
        shutil.copy(os.path.join('..', 'test', 'Makefile.Microsoft_nmake'),
                    os.path.join(test_dir, 'test', 'Makefile.Microsoft_nmake'))
        # Copy directories with support files
        for d in ['common', 'test_common', 'crypto_sign', 'crypto_kem']:
            shutil.copytree(
                os.path.join('..', 'test', d),
                os.path.join(test_dir, 'test', d)
            )

        shutil.copytree(impl_path, new_impl_dir)

    def destructor():
        """Clean up the isolated files"""
        shutil.rmtree(test_dir, ignore_errors=True)

    return (test_dir, new_impl_dir, initializer, destructor)


def run_subprocess(command, working_dir='.', env=None, expected_returncode=0,
                   print_output=True):
    """
    Helper function to run a shell command and report success/failure
    depending on the exit status of the shell command.
    """
    if env is not None:
        env_ = os.environ.copy()
        env_.update(env)
        env = env_

    # Note we need to capture stdout/stderr from the subprocess,
    # then print it, which the unittest will then capture and
    # buffer appropriately
    print(working_dir + " > " + " ".join(command))
    result = subprocess.run(
        command,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        cwd=working_dir,
        env=env,
    )
    if print_output:
        print(result.stdout.decode('utf-8'))
    if expected_returncode is not None:
        assert result.returncode == expected_returncode, \
            "Got unexpected return code {}".format(result.returncode)
    else:
        return (result.returncode, result.stdout.decode('utf-8'))
    return result.stdout.decode('utf-8')


def make(*args, working_dir='.', env=None, expected_returncode=0, **kwargs):
    """
    Runs a make target in the specified working directory

    Usage:
        make('clean', 'targetb', SCHEME='bla')
    """
    if os.name == 'nt':
        make_command = ['nmake', '/f', 'Makefile.Microsoft_nmake',
                        '/NOLOGO', '/E']
        # we need SCHEME_UPPERCASE and IMPLEMENTATION_UPPERCASE with nmake
        for envvar in ['IMPLEMENTATION', 'SCHEME']:
            if envvar in kwargs:
                kwargs['{}_UPPERCASE'.format(envvar)] = (
                        kwargs[envvar].upper().replace('-', ''))
    else:
        make_command = ['make']

    return run_subprocess(
        [
            *make_command,
            *['{}={}'.format(k, v) for k, v in kwargs.items()],
            *args,
        ],
        working_dir=working_dir,
        env=env,
        expected_returncode=expected_returncode,
    )


def skip_windows(message="This test is not supported on Windows"):
    def wrapper(f):
        @functools.wraps(f)
        def skip_windows(*args, **kwargs):
            raise unittest.SkipTest(message)
        if os.name == 'nt':
            return skip_windows
        else:
            return f
    return wrapper


@lru_cache(maxsize=None)
def ensure_available(executable):
    """
    Checks if a command is available.

    If a command MUST be available, because we are in a CI environment,
    raises an AssertionError.

    In the docker containers, on Travis and on Windows, CI=true is set.
    """
    path = shutil.which(executable)
    if path:
        return path

    # Installing clang-tidy on LLVM will be too much of a mess.
    if ((executable == 'clang-tidy' and sys.platform == 'darwin')
            or 'CI' not in os.environ):
        raise unittest.SkipTest(
                "{} is not available on PATH. Install it to run this test.{}"
                .format(executable, "" if not os.name == 'nt'
                        else "On Windows, make sure to add it to PATH")
                )
    raise AssertionError("{} not available on CI".format(executable))


def permit_test(testname, *args, **kwargs):
    if len(args) == 0:
        thing = list(kwargs.values())[0]
    else:
        thing = args[0]

    if 'PQCLEAN_ONLY_TESTS' in os.environ:
        if not(testname.lower() in os.environ['PQCLEAN_ONLY_TESTS'].lower().split(',')):
            return False
    if 'PQCLEAN_SKIP_TESTS' in os.environ:
        if testname.lower() in os.environ['PQCLEAN_SKIP_TESTS'].lower().split(','):
            return False

    if isinstance(thing, pqclean.Implementation):
        scheme = thing.scheme
    elif isinstance(thing, pqclean.Scheme):
        scheme = thing
    else:
        return True

    if 'PQCLEAN_ONLY_TYPES' in os.environ:
        if not(scheme.type.lower() in os.environ['PQCLEAN_ONLY_TYPES'].lower().split(',')):
            return False
    if 'PQCLEAN_SKIP_TYPES' in os.environ:
        if scheme.type.lower() in os.environ['PQCLEAN_SKIP_TYPES'].lower().split(','):
            return False
    if 'PQCLEAN_ONLY_SCHEMES' in os.environ:
        if not(scheme.name.lower() in os.environ['PQCLEAN_ONLY_SCHEMES'].lower().split(',')):
            return False
    if 'PQCLEAN_SKIP_SCHEMES' in os.environ:
        if scheme.name.lower() in os.environ['PQCLEAN_SKIP_SCHEMES'].lower().split(','):
            return False

    if 'PQCLEAN_ONLY_DIFF' in os.environ:
        if shutil.which('git') is not None:
            # if we're on a non-master branch, and the only changes are in schemes,
            # only run tests on those schemes
            branch_result = subprocess.run(
                ['git', 'status', '--porcelain=2', '--branch'],
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                cwd="..",
            )
            # ensure we're in a working directory
            if branch_result.returncode != 0:
                return True
            # ensure we're not on master branch
            for branch_line in branch_result.stdout.decode('utf-8').splitlines():
                tokens = branch_line.split(' ')
                if tokens[0] == '#' and tokens[1] == 'branch.head':
                    if tokens[2] == 'master':
                        return True
            # where are there changes?
            diff_result = subprocess.run(
                ['git', 'diff', '--name-only', 'origin/master'],
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT
            )
            assert diff_result.returncode == 0, \
                "Got unexpected return code {}".format(diff_result.returncode)
            for diff_line in diff_result.stdout.decode('utf-8').splitlines():
                # Git still returns UNIX-style paths on Windows, normalize
                diff_line = os.path.normpath(diff_line)
                # don't skip test if there are any changes outside schemes
                if (not diff_line.startswith('crypto_kem') and
                        not diff_line.startswith('crypto_sign') and
                        not diff_line.startswith(os.path.join('test', 'duplicate_consistency'))):
                    logging.info("Running all tests as there are changes "
                                 "outside of schemes")
                    return True
                # do test if the scheme in question has been changed
                if diff_line.startswith(thing.path(base='')):
                    return True
                # do test if the scheme's duplicate_consistency files have been changed
                if diff_line.startswith(os.path.join('test', 'duplicate_consistency', scheme.name.lower())):
                    return True
            # there were no changes outside schemes, and the scheme in question had no diffs
            return False

    return True


def filtered_test(func):
    funcname = func.__name__[len("test_"):]

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if permit_test(funcname, *args, **kwargs):
            return func(*args, **kwargs)
        else:
            raise unittest.SkipTest("Test disabled by filter")
    return wrapper


@lru_cache(maxsize=1)
def get_cpu_info():
    the_info = None
    while the_info is None or 'flags' not in the_info:
        import cpuinfo
        the_info = cpuinfo.get_cpu_info()

    return the_info



import json
import os

import pytest

import helpers
import pqclean


@pytest.mark.parametrize(
    'implementation,test_dir,impl_path, init, destr',
    [(impl, *helpers.isolate_test_files(impl.path(), 'test_printparams_'))
     for impl in pqclean.Scheme.all_implementations()],
    ids=[str(impl) for impl in pqclean.Scheme.all_implementations()],
)
@helpers.filtered_test
def test_metadata_sizes(implementation, impl_path, test_dir, init, destr):
    init()
    metadata = implementation.scheme.metadata()
    dest_dir = os.path.join(test_dir, 'bin')
    helpers.make('printparams',
                 TYPE=implementation.scheme.type,
                 SCHEME=implementation.scheme.name,
                 IMPLEMENTATION=implementation.name,
                 SCHEME_DIR=impl_path,
                 DEST_DIR=dest_dir,
                 working_dir=os.path.join(test_dir, 'test'))

    out = helpers.run_subprocess(
        [os.path.join(dest_dir, 'printparams_{}_{}{}'.format(
            implementation.scheme.name,
            implementation.name,
            '.exe' if os.name == 'nt' else ''
        ))]
    ).replace('\r', '')

    parsed = json.loads(out)

    assert parsed['CRYPTO_SECRETKEYBYTES'] == metadata['length-secret-key']
    assert parsed['CRYPTO_PUBLICKEYBYTES'] == metadata['length-public-key']
    assert parsed['CRYPTO_ALGNAME'] == metadata['name']

    if implementation.scheme.type == 'kem':
        assert (
            parsed['CRYPTO_CIPHERTEXTBYTES'] == metadata['length-ciphertext'])
        assert parsed['CRYPTO_BYTES'] == metadata['length-shared-secret']
    else:
        assert parsed['CRYPTO_BYTES'] == metadata['length-signature']
    destr()


if __name__ == '__main__':
    import sys
    pytest.main(sys.argv)



"""
Checks that a LICENSE or LICENSE.txt file is present for every
implementation of the specified scheme.
"""

import os

import pytest

import helpers
import pqclean


@pytest.mark.parametrize(
    'implementation',
    pqclean.Scheme.all_implementations(),
    ids=str,
)
@helpers.filtered_test
def test_license(implementation):
    p1 = os.path.join(implementation.path(), 'LICENSE')
    p2 = os.path.join(implementation.path(), 'LICENSE.txt')
    assert(os.path.isfile(p1) or os.path.isfile(p2))


if __name__ == '__main__':
    import sys
    pytest.main(sys.argv)



"""
Runs functional tests for common crypto functions (e.g., fips202, sha2, aes).

Doesn't currently need isolation for parallelisation
"""

import os
import re

import helpers


def pytest_generate_tests(metafunc):
    argvalues = []
    for d in os.listdir('test_common'):
        primitive = re.sub(r"\.c$", "", d)
        argvalues.append(primitive)
    metafunc.parametrize('primitive', argvalues)


@helpers.skip_windows()
@helpers.filtered_test
def test_common(primitive):
    binname = os.path.join('..', 'bin', 'test_common_'+primitive)
    helpers.make(binname)
    helpers.run_subprocess([binname])


if __name__ == '__main__':
    import pytest
    import sys
    pytest.main(sys.argv)



"""
Checks that the archive library can be successfully built for every
scheme/implementation.
"""
import pytest

import helpers
import pqclean


@pytest.mark.parametrize(
    'implementation,test_dir,impl_dir, init, destr',
    [(impl, *helpers.isolate_test_files(impl.path(), 'test_functest_'))
     for impl in pqclean.Scheme.all_supported_implementations()],
    ids=[str(impl) for impl in pqclean.Scheme.all_supported_implementations()],
)
@helpers.filtered_test
def test_compile_lib(implementation, test_dir, impl_dir, init, destr):
    init()
    helpers.make('clean', working_dir=impl_dir)
    helpers.make(working_dir=impl_dir)
    destr()


if __name__ == '__main__':
    import sys
    pytest.main(sys.argv)



"""
Checks that no dynamic memory functions are used
"""

import pytest

import helpers
import pqclean


@pytest.mark.parametrize(
    'implementation,test_dir,impl_path, init, destr',
    [(impl, *helpers.isolate_test_files(impl.path(), 'test_functest_'))
     for impl in pqclean.Scheme.all_supported_implementations()],
    ids=[str(impl) for impl in pqclean.Scheme.all_supported_implementations()],
)
@helpers.skip_windows()
@helpers.filtered_test
def test_dynamic_memory(implementation, test_dir, impl_path, init, destr):
    init()
    # 'make' will take care of not rebuilding existing library files
    helpers.make(working_dir=impl_path)
    scheme_name = implementation.scheme.name
    out = helpers.run_subprocess(
        ['nm', '-g', 'lib{}_{}.a'.format(scheme_name, implementation.name)],
        impl_path,
    )

    lines = out.strip().split("\n")

    for line in lines:
        for function in ['malloc', 'free', 'realloc', 'calloc']:
            if line.endswith('U {}'.format(function)):
                raise AssertionError(
                    "Illegal use of dynamic memory function "
                    "'{function}'".format(function=function))

    destr()


if __name__ == '__main__':
    import sys
    pytest.main(sys.argv)
